/************************************************************************
 *									*
 *			Copyright (c) 1990 by     			*
 *		Digital Equipment Corporation, Maynard, MA		*
 *			All rights reserved.				*
 *									*
 *   This software is furnished under a license and may be used and	*
 *   copied  only  in accordance with the terms of such license and	*
 *   with the  inclusion  of  the  above  copyright  notice.   This	*
 *   software  or  any  other copies thereof may not be provided or	*
 *   otherwise made available to any other person.  No title to and	*
 *   ownership of the software is hereby transferred.			*
 *									*
 *   The information in this software is subject to change  without	*
 *   notice  and should not be construed as a commitment by Digital	*
 *   Equipment Corporation.						*
 *									*
 *   Digital assumes no responsibility for the use  or  reliability	*
 *   of its software on equipment which is not supplied by Digital.	*
 *									*
 *									*
 ************************************************************************
 *									*
 *			Modification History				*
 *									*
 ************************************************************************
 *									*
 * Authors:								*
 *	DBW - David B. Wecker, Database Systems Research Group	      	*
 *		DTN: 	523-2794					*
 *		PHONE:	719/260-2794					*
 *		ENET:	CHILD::WECKER					*
 *		USENET:	decwrl!child.dec.com!wecker			*
 *		ARPA:	wecker%child.dec.com@decwrl.dec.com		*
 *		MS: 	CXN/2 Upper - 564				*
 *		MAIL:	1175 Chapel Hills Drive, Col Spgs, CO 80920-3952*
 *									*
 * Versions:								*
 *  v1.0 DBW 871105 Ultrix/X10 based version				*
 *  v2.1 DBW 871117 Major re-write (non X based)			*
 *  v2.2 DBW 880114 Update of internal functions			*
 *  v3.0 DBW 881213 Major new release (replaced NeurDS compiler)	*
 *  v3.1 DBW 890501 Minor release for external use (portability fixes)	*
 *									*
 ************************************************************************/


/*************************************************************************
Glass-Mackey eqn:


	.     a x(t - tau)
	x = -------------------  - b x(t)
		   10
	      1 + x (t - tau)


	      a x(t - tau)
    x(t+1) = ----------------  + ( 1 - b ) x(t)
	           10
	      1 + x (t - tau)


Typical values:

	a 		= .2
	b 		= .1
	tau		= 30	(fractal attractor of dimension 3.5)
	initial x's	= .5	(constant over strip of width tau)
	delta		=  6	
	P		=  6

Output (ranges from 0 to 1.35 (approx) scaled to [0 .. 1]):

	O		x(t + P)
	I0		x(t)
	I1		x(t - 1 * delta)
	I2		x(t - 2 * delta)
	I3		x(t - 3 * delta)
	I4		x(t - 4 * delta)

************************************************************************/

#include "stdio.h"
#include "math.h"

#define	    A	    .2
#define	    B	    .1
#define	    INIVAL  .5
#define	    DELTA   6
int	    TAU	=   30;
int	    P =	    6;

#define	    COUNT   1000
#define	    BUFMAX	1000
#define	    IDX(a)	((a) % BUFMAX)
#define	    SCALE(a)	((a) / 1.35)

double	buf[BUFMAX];

doval(t) {
    int	    cur,prv,tau;

    cur		= IDX(t);
    prv		= IDX(t-1);
    tau		= IDX(t-TAU);
    buf[cur]    = buf[prv] * (1.0 - B);
    buf[cur]   += (A * buf[tau]) / (1 + pow(buf[tau],10.0));
    }

main(argc,argv)
char	**argv;
    {
    int	    i,t;
    FILE    *fp;

    if (argc > 1)   P   = atoi(argv[1]);
    if (argc > 2)   TAU = atoi(argv[2]);

    fp = fopen("GlassMackey.dat","w");
    if (!fp) {
	printf("Can't open output file\n");
	exit(1);
	}

    for (t=0; t<BUFMAX; t++) buf[t] = INIVAL;
    for (t=TAU; t<BUFMAX; t++) doval(t);
    for (; t<BUFMAX+COUNT; t++) {
	doval(t);
	fprintf(fp,"%9.6f",SCALE(buf[IDX(t)]));
	for (i=0; i<5; i++) 
	    fprintf(fp," %9.6f",SCALE(buf[IDX(t-(P+i*DELTA))]));
	fprintf(fp,"\n");
	}
    fclose(fp);
    }
