/************************************************************************
 *									*
 *			Copyright (c) 1990 by     			*
 *		Digital Equipment Corporation, Maynard, MA		*
 *			All rights reserved.				*
 *									*
 *   This software is furnished under a license and may be used and	*
 *   copied  only  in accordance with the terms of such license and	*
 *   with the  inclusion  of  the  above  copyright  notice.   This	*
 *   software  or  any  other copies thereof may not be provided or	*
 *   otherwise made available to any other person.  No title to and	*
 *   ownership of the software is hereby transferred.			*
 *									*
 *   The information in this software is subject to change  without	*
 *   notice  and should not be construed as a commitment by Digital	*
 *   Equipment Corporation.						*
 *									*
 *   Digital assumes no responsibility for the use  or  reliability	*
 *   of its software on equipment which is not supplied by Digital.	*
 *									*
 *									*
 ************************************************************************
 *									*
 *			Modification History				*
 *									*
 ************************************************************************
 *									*
 * Authors:								*
 *	DBW - David B. Wecker, Database Systems Research Group	      	*
 *		DTN: 	523-2794					*
 *		PHONE:	719/260-2794					*
 *		ENET:	CHILD::WECKER					*
 *		USENET:	decwrl!child.dec.com!wecker			*
 *		ARPA:	wecker%child.dec.com@decwrl.dec.com		*
 *		MS: 	CXN/2 Upper - 564				*
 *		MAIL:	1175 Chapel Hills Drive, Col Spgs, CO 80920-3952*
 *									*
 * Versions:								*
 *  v1.0 DBW 871105 Ultrix/X10 based version				*
 *  v2.1 DBW 871117 Major re-write (non X based)			*
 *  v2.2 DBW 880114 Update of internal functions			*
 *  v3.0 DBW 881213 Major new release (replaced NeurDS compiler)	*
 *  v3.1 DBW 890501 Minor release for external use (portability fixes)	*
 *									*
 ************************************************************************/


/************************************************************************
 *
 * Code to do Second order BACKPROPAGATION under NeurDS v3.1
 *
 *  Defined fields are:
 *	UNITS:
 *		sum	- input sum to neuron
 *		val	- output value
 *		err	- error for backprop
 *		csum	- change in input sum
 *		cval	- change in output value
 *		cerr	- change in error
 *		levid	- level (low 16), id (high 16)
 *	LINKS:
 *		weight	- current weight on link
 *		cweight	- change in weight value
 *
 *	Parameters:
 *		A1	= learn rate
 *		A2	= result tolerance
 *		A3	= exp factor
 *		A4	= momentum
 *		A5	= 2nd order factor
 *
 ************************************************************************/

#include "stdio.h"
#include "math.h"
#include "NeurDS.h"

double	aer,aep,awr,awp;
int	nr,np;

/* SIM_INIT - perform any model specific initialization 	  */

SIM_INIT() {
    SVERSION =
    "2nd Order Back-Propagation Simulator version 3.1 (DBW 890501) -----------";
    PLOTCHAR	= "EW";
    MINVAL	= 0;
    CONTINUOUS	= 0;
    }

/* SIM_RAND	- randomize networks					*/
SIM_RAND() {
    randomize(1,INPUT,OUTPUT-1,0.45,0.1,-1.0,2.0);
    }

/* SIM_LINIT -	perform any initializations before learning		*/

SIM_LINIT() {
    register    int	net;

    aer	    = 0.0;
    awr	    = 0.0;
    aep	    = 0.0;
    awp	    = 0.0;
    nr	    = 0;
    np	    = 0;
    for (net = HIDDEN; net <= OUTPUT; net++)
	for (NC_u1 = networks[net]; NC_u1; NC_u1 = NC_u1->next)
	    for (NC_l1=NC_u1->inp; NC_l1; NC_l1=NC_l1->inpnxt)
	    NC_u1->cval = NC_l1->cweight = 0.0;
    }

/* SIM_RUN -	Run the model (forward pass)				*/

SIM_RUN(mode) {
    register	int	net,retval;

    for (net = HIDDEN; net <= OUTPUT; net++) {
	for (NC_u1 = networks[net]; NC_u1; NC_u1 = NC_u1->next) NC_u1->sum = NC_u1->csum  = 0.0;
	for (NC_u1 = networks[net]; NC_u1; NC_u1 = NC_u1->next)
	    for (NC_l1=NC_u1->inp; NC_l1; NC_l1=NC_l1->inpnxt)   {
	    NC_u1->sum  += NC_l1->weight  * NC_l1->inp->val;
	    NC_u1->csum += NC_l1->weight  * NC_l1->inp->cval + A5 * NC_l1->cweight * NC_l1->inp->val;
	    }
	for (NC_u1 = networks[net]; NC_u1; NC_u1 = NC_u1->next) {
	    NC_u1->val  = 1.0 / (1.0 + exp(-A3 * NC_u1->sum));
	    NC_u1->sum  = NC_u1->val * (1.0 - NC_u1->val);
	    NC_u1->cval = NC_u1->sum * NC_u1->csum;
	    }
	}
    retval = 1;
    OUTERR = 0.0;
    for (NC_u1 = networks[TARGET], NC_u2 = networks[OUTPUT];
	    NC_u1 && NC_u2;
	    NC_u1 = NC_u1->next,NC_u2 = NC_u2->next) {
	NC_u2->err  = NC_u1->val - NC_u2->val;
	if (NC_u2->err > A2 || NC_u2->err < -A2) retval = 0;
	OUTERR  += NC_u2->err * NC_u2->err;
	NC_u2->cerr   = -2.0 * NC_u2->cval;
	NC_u2->err	*= 2.0;
	NC_u2->csum	 = NC_u2->sum * (NC_u2->cerr + NC_u2->err * (1.0 - 2.0 * NC_u2->val) * NC_u2->csum);
	NC_u2->sum   *= NC_u2->err;
	}
    if (mode == 0 || mode == 1) {
	aer += OUTERR;
	aep += OUTERR;
	nr++;
	np++;
	if (!retval) {
	    awr += 1.0;
	    awp += 1.0;
	    }
	}
    return retval;
    }


/* SIM_LEARN	- backward pass to fix up weights			*/

SIM_LEARN(curbad) {
    register    net;

    for (net = OUTPUT-1; net >= INPUT; net--) {
	for (NC_u1 = networks[net]; NC_u1; NC_u1 = NC_u1->next) NC_u1->err  = NC_u1->cerr = 0.0;
	for (NC_u1 = networks[net]; NC_u1; NC_u1 = NC_u1->next)
	    for (NC_l1=NC_u1->out; NC_l1; NC_l1=NC_l1->outnxt) {
	    NC_u1->err	+= NC_l1->out->sum  * NC_l1->weight;
	    NC_u1->cerr	+= NC_l1->out->csum * NC_l1->weight + A5 * NC_l1->cweight * NC_l1->out->sum ;
	    NC_l1->cweight	 = A4 * NC_l1->cweight
			    + (A1 * NC_l1->out->sum + NC_l1->out->csum) * NC_u1->val
			    + NC_l1->out->sum * NC_u1->cval;
	    NC_l1->weight	+= NC_l1->cweight;
	    }
	for (NC_u1 = networks[net]; NC_u1; NC_u1 = NC_u1->next) {
	    NC_u1->csum  = NC_u1->sum * (NC_u1->cerr + NC_u1->err * (1.0 - 2.0 * NC_u1->val) * NC_u1->csum);
	    NC_u1->sum  *= NC_u1->err;
	    }
	}
    }

/* SIM_LPASS	- end of one pass of learning				*/
SIM_LPASS(loop,maxassoc) {
    }

/* SIM_LEND	- end of learning report				*/
SIM_LEND(loop) {
    PR(";Total loops = %d\n",loop);
    }

/* SIM_REPORT	- give info during run					*/
double SIM_REPORT(loop,numassoc) {
    double retval;

    if (nr > 0) {
	aer	 = sqrt(aer / (double)nr);
	awr	/= (double)nr / (double)numassoc;
	}

    PR(";\tLoop %5d: Wrong %7.3f Err %8.4lf\n",
	loop,awr,aer);

    retval  = aer;
    aer	    = 0.0;
    awr	    = 0.0;
    nr      = 0;

    return retval;
    }

/* SIM_PLOT	- give info during run					*/
SIM_PLOT(numassoc) {
    if (np > 0) {
	aep  = sqrt(aep / (double)np);
	awp /= (double)np / (double)numassoc;
	}
    plot(0,aep);
    plot(1,awp);
    np	= 0;
    aep = 0.0;
    awp = 0.0;
    }

/* SIM_PAT	- hand a pattern back to the simulator			*/

/* mode 0 = run, 1 = learn, 2 = run/nostats, 3 = dump patterns */
SIM_PAT(index,mode) {
    do_pat(index);
    }


