/************************************************************************
 *									*
 *			Copyright (c) 1990 by     			*
 *		Digital Equipment Corporation, Maynard, MA		*
 *			All rights reserved.				*
 *									*
 *   This software is furnished under a license and may be used and	*
 *   copied  only  in accordance with the terms of such license and	*
 *   with the  inclusion  of  the  above  copyright  notice.   This	*
 *   software  or  any  other copies thereof may not be provided or	*
 *   otherwise made available to any other person.  No title to and	*
 *   ownership of the software is hereby transferred.			*
 *									*
 *   The information in this software is subject to change  without	*
 *   notice  and should not be construed as a commitment by Digital	*
 *   Equipment Corporation.						*
 *									*
 *   Digital assumes no responsibility for the use  or  reliability	*
 *   of its software on equipment which is not supplied by Digital.	*
 *									*
 *									*
 ************************************************************************
 *									*
 *			Modification History				*
 *									*
 ************************************************************************
 *									*
 * Authors:								*
 *	DBW - David B. Wecker, Database Systems Research Group	      	*
 *		DTN: 	523-2794					*
 *		PHONE:	719/260-2794					*
 *		ENET:	CHILD::WECKER					*
 *		USENET:	decwrl!child.dec.com!wecker			*
 *		ARPA:	wecker%child.dec.com@decwrl.dec.com		*
 *		MS: 	CXN/2 Upper - 564				*
 *		MAIL:	1175 Chapel Hills Drive, Col Spgs, CO 80920-3952*
 *									*
 * Versions:								*
 *  v1.0 DBW 871105 Ultrix/X10 based version				*
 *  v2.1 DBW 871117 Major re-write (non X based)			*
 *  v2.2 DBW 880114 Update of internal functions			*
 *  v3.0 DBW 881213 Major new release (replaced NeurDS compiler)	*
 *  v3.1 DBW 890501 Minor release for external use (portability fixes)	*
 *									*
 ************************************************************************/


/**********************************************************************
 *
 * Tiny pseudo "curses" package (runs on U__X, VMS, or MCH_AMIGA)
 *
 *	v1.0	870117	DBW - D. Wecker, initial hack
 *	  .1	880117	DBW - Added all sorts of goodies
 *
 *
 *  Routines:
 *	initscr(lev,"scrn nam")	- start up system (lev=0 says no tty)
 *	endwin()		- close down the system
 *
 *	move(r,c)		- move to row and col (23x78 on Amiga)
 *	clrtoeol()		- clear to end of line
 *	clrtobot()		- clear to end of screen
 *	clear()			- clear whole screen (move to 0,0)
 *	standout()		- start bold
 *	standend()		- end bold
 *	addstr(s)		- add string to screen
 *	addchr(c)		- add character to screen
 *	printw(fmt,arg1,...)	- formatted screen output
 *	refresh()		- update the screen
 *	resize()		- resize screen for new rows and columns
 *	inch()			- read a character off of the screen
 *	inrow(m,str)		- return row as a string (m=1 do BOLDs)
 *	scroll(r1,r2)		- scroll region from r1 to r2
 *	spush()			- save a copy of the current screen
 *	spop()			- restore previous screen
 *
 *	touchwin()		- force full repaint
 *	ttgetc()		- read one key from keyboard
 *	ttgets(str)		- read/echo one line from terminal
 *	ttputc()		- write char   to hardware screen
 *	ttputs()		- write string to hardware screen
 *	ttflush()		- flush output to hardware screen
 *
 *
 *	tabify(s,d)		- take string s and tabify it into d
 *
 **********************************************************************/

#ifdef VMS
#include <stdio.h>
#include <stsdef.h>
#include <ssdef.h>
#include <descrip.h>
#include <iodef.h>
#include <ttdef.h>
#include <tt2def.h>
#include <varargs.h>

#define NIBUF   1		/* Input buffer size */
#define NOBUF   1024		/* MM says bug buffers win! */
#define EFN     0		/* Event flag */

char obuf[NOBUF];		/* Output buffer */
int nobuf;			/* # of bytes in above */
int oldmode[3];			/* Old TTY mode bits */
int newmode[3];			/* New TTY mode bits */
short iochan;			/* TTY I/O channel */
struct dsc$descriptor  idsc;
struct dsc$descriptor  odsc;
char oname[40];
int iosb[2];
int term[2];
int status;
#endif

#ifdef MCH_AMIGA
extern	char		*Open();
extern	long		Read(),Write();
extern	void		Close();
#define NEW		1006L
#define AMG_MAXBUF	1024
static char		*terminal = 0L;
static char		scrn_tmp[AMG_MAXBUF+1];
static long		scrn_tmp_p = 0L;
static char		scrnam[128];
#ifndef NULL
#define	NULL	((void *)0)
#endif
#endif

#ifdef U__X
#include <sys/ioctl.h>
#include <stdio.h>
#include <varargs.h>

#ifdef hpux
#include <termio.h>
struct termio old_tty,new_tty;
#else
#include <sgtty.h>
struct sgttyb old_tty,new_tty;
#endif /* hpux */
#endif

#ifdef MCH_AMIGA
#define MROWS	47
#define	MCOLS	78
int	ROWS =  23;
int	COLS =  78;

#else
#define MROWS	66
#define	MCOLS	132
int	ROWS =  24;
int	COLS =  80;
#endif

#define NORMAL	0x00
#define BOLD	0x80

typedef char	SCRN[MROWS][MCOLS];

SCRN	nscrn,cscrn,*pscrn[10];
int	row,
	col,
	mode,
	scroll1,
	scroll2;
char	c_str[256],c_str2[256];
short	plev	= -1;
short	inited	= 0;

move(y,x)
int y,x;
    {
    row = y;
    col = x;
    }

clrtoeol() {
    int i;

    for (i = col; i < COLS; i++) nscrn[row][i] = ' ' | mode;
    }

printw(fmt, va_alist)
char	*fmt;
va_dcl
    {
    va_list	    ap;
    char	    str[512];
    FILE	    strbuf;

    va_start(ap);
    vsprintf(str,fmt,ap);
    addstr(str);
    }

addchr(c)
char	c;
    {
    char    str[2];

    str[0] = c;
    str[1] = '\0';
    addstr(str);
    }

addstr(pstr)
char *pstr;
    {
    int	    i,j,r1,c1;

    for (j = 0; col < COLS && pstr[j] != '\000'; j++) {
	if (pstr[j] == '\n') {
	    if (row == scroll2) {
		refresh();
		sprintf(c_str,"\033[%d;%dr\033[%d;1H\n\033[1;99r",
		    scroll1+1,scroll2+1,scroll2+1);
		ttputs(c_str);
		ttflush();
		for (r1=scroll1; r1<scroll2; r1++)
		    for (c1=0; c1 < COLS; c1++) {
			nscrn[r1][c1] = nscrn[r1+1][c1];
			cscrn[r1][c1] = cscrn[r1+1][c1];
			}
		for (c1=0; c1 < COLS; c1++) {
		    nscrn[scroll2][c1] = ' ' | mode;
		    cscrn[scroll2][c1] = ' ' | mode;
		    }
		}
	    else if (row != ROWS-1) row++;
	    col = 0;
	    clrtoeol();
	    }
	else if (pstr[j] < ' ')	    nscrn[row][col++] = ' '	| mode;
	else			    nscrn[row][col++] = pstr[j] | mode;
	}
    }

scroll(r1,r2) {
    scroll1 = r1;
    scroll2 = r2;
    }

clrtobot() {
    int i,j;

    clrtoeol();
    for (i = row+1; i < ROWS; i++)
	for (j = 0; j < COLS; j++)
	    nscrn[i][j] = ' ' | mode;
    }

standout() {
    mode = BOLD;
    }

standend() {
    mode = NORMAL;
    }

initscr(lev,nam)
char	*nam;
    {
    int	    i,j;
    char    title[256];

    row	    = 0;
    col	    = 0;
    scroll1 = 0;
    scroll2 = ROWS-1;
    mode    = NORMAL;
    for (i = 0; i < MROWS; i++)
	for (j = 0; j < MCOLS; j++)
	    nscrn[i][j] = cscrn[i][j] = ' ';
    row	    = 0;
    col	    = 0;
    scroll1 = 0;
    scroll2 = ROWS-1;
    mode    = NORMAL;
    inited  = lev;
    for (i = 0; i < MROWS; i++)
	for (j = 0; j < MCOLS; j++)
	    nscrn[i][j] = cscrn[i][j] = ' ';
    if (lev == 0) return;

#ifdef MCH_AMIGA
    if (nam != NULL) strcpy(scrnam,nam);
    else	     strcpy(scrnam,"Curses Screen");
    sprintf(title,"RAW:1/1/639/199/%s",scrnam);
    terminal = Open(title,(long)NEW);
#endif
#ifdef VMS
    odsc.dsc$a_pointer = "TT";
    odsc.dsc$w_length = strlen(odsc.dsc$a_pointer);
    odsc.dsc$b_dtype = DSC$K_DTYPE_T;
    odsc.dsc$b_class = DSC$K_CLASS_S;
    idsc.dsc$b_dtype = DSC$K_DTYPE_T;
    idsc.dsc$b_class = DSC$K_CLASS_S;
    do {
	idsc.dsc$a_pointer = odsc.dsc$a_pointer;
	idsc.dsc$w_length = odsc.dsc$w_length;
	odsc.dsc$a_pointer = &oname[0];
	odsc.dsc$w_length = sizeof(oname);
	status = LIB$SYS_TRNLOG(&idsc, &odsc.dsc$w_length, &odsc);
	if (status!=SS$_NORMAL && status!=SS$_NOTRAN) exit(status);
	if (oname[0] == 0x1B) {
	    odsc.dsc$a_pointer += 4;
	    odsc.dsc$w_length -= 4;
	    }
	}
    while (status == SS$_NORMAL);
    status = SYS$ASSIGN(&odsc, &iochan, 0, 0);
    if (status != SS$_NORMAL) exit(status);
#endif

#ifdef U__X
#ifdef hpux
	ioctl( 0, TCGETA, &old_tty );
	ioctl( 0, TCGETA, &new_tty );
	new_tty.c_lflag &= ~ICANON;
	new_tty.c_lflag &= ~ECHO;
	new_tty.c_cc[VMIN] = 1;
	new_tty.c_cc[VTIME] = 0;
	ioctl( 0, TCSETA, &new_tty );
#else
    ioctl(0,TIOCGETP,&old_tty);
    ioctl(0,TIOCGETP,&new_tty);
    new_tty.sg_flags |= CBREAK;
    new_tty.sg_flags &= ~ECHO;
    ioctl(0,TIOCSETP,&new_tty);
#endif /* hpux */
#endif

    ttputs("\033[2J");
    }

clear() {
    row = 0;
    col = 0;
    clrtobot();
    }

spush() {
    int	    r,c;
    char    *malloc();

    if (plev == 9) return;
    pscrn[++plev] = (SCRN *)malloc(sizeof(SCRN));
    for (r=0; r<MROWS; r++)
	for (c=0; c<MCOLS; c++)
	    (*pscrn[plev])[r][c] = nscrn[r][c];
    }

spop() {
    int	r,c;

    if (plev == -1) return;
    for (r=0; r<MROWS; r++)
	for (c=0; c<MCOLS; c++)
	    nscrn[r][c] = (*pscrn[plev])[r][c];
    free(pscrn[plev--]);
    }

endwin() {
    if (!inited) return;

    move(ROWS-1,0);
    refresh();
    ttputs("\033[r\033[99H\n");
    ttflush();

#ifdef MCH_AMIGA
    amg_flush();
    Close(terminal);
#endif

#ifdef VMS
    status = SYS$DASSGN(iochan);
    if (status != SS$_NORMAL) exit(status);
#endif

#ifdef U__X
#ifdef hpux
	ioctl( 0, TCSETA, &old_tty );
#else
    ioctl(0,TIOCSETP,&old_tty);
#endif /* hpux */
#endif

    }

resize(r,c)
int r,c;
    {
    char    title[256];

    if (r > MROWS) r = MROWS;
    if (c > MCOLS) c = MCOLS;
    ROWS = r;
    COLS = c;
    scroll1 = 0;
    scroll2 = ROWS-1;
    touchwin();
    if (!inited) return;
#ifdef MCH_AMIGA
    amg_flush();
    Close(terminal);
    if (r > 23) sprintf(title,"RAW:1/1/639/399/%s",scrnam);
    else	sprintf(title,"RAW:1/1/639/199/%s",scrnam);
    terminal = Open(title,(long)NEW);
#endif
    refresh();
    }

char *inrow(dom,str)
char	*str;
    {
    int	chr,i,j,m;

    m = NORMAL;
    for (i=0,j=0; j<COLS; j++) {
	chr=nscrn[row][j];
	if (dom && (chr & BOLD) != m) {
	    str[i++] = '\033';
	    str[i++] = '[';
	    if (m == NORMAL) { str[i++] = '7'; m = BOLD; }
	    else	     m = NORMAL;
	    str[i++] = 'm';
	    }
	str[i++] = chr & 0x7F;
	}
    for (i--; i >= 0 && str[i] == ' '; i--) ;
    str[++i] = '\n';
    str[++i] = '\000';
    return str;
    }

char inch() {
    return(nscrn[row][col] & 0x7F);
    }

touchwin() {
    int i,j;

    for (i=0; i<MROWS; i++)
	for (j=0; j<MCOLS; j++)
	    cscrn[i][j] = ' ';
    ttputs("\033[H\033[2J");
    }

refresh() {
    int	i,j,mode,curpos;

    if (!inited) return;
    mode = NORMAL;
    for (i=0; i < ROWS; i++) {
	curpos = -1;
	for (j = 0; j < COLS; j++) {
	    if (nscrn[i][j] != cscrn[i][j]) {
		if (curpos == -1) {
		    sprintf(c_str,"\033[%d;%dH",i+1,j+1);
		    ttputs(c_str);
		    curpos = j;
		    }
		else {
		    sprintf(c_str,"\033[%dC",j-curpos);
		    ttputs(c_str);
		    curpos = j;
		    }
		while (nscrn[i][j] != cscrn[i][j]) {
		    if (mode == NORMAL && (nscrn[i][j] & BOLD) == BOLD) {
			ttputs("\033[7m");
			mode = BOLD;
			}
		    else if (mode == BOLD && (nscrn[i][j] & BOLD) == NORMAL) {
			ttputs("\033[0m");
			mode = NORMAL;
			}
		    cscrn[i][j] = nscrn[i][j];
		    ttputc(nscrn[i][j] & 0x7F);
		    curpos++;
		    j++;
		    }
		}
	    }
	}
    sprintf(c_str,"\033[%d;%dH",row+1,col+1);
    ttputs(c_str);
    if (mode) ttputs("\033[0m");
    ttflush();
    }

char *ttgets(str)
char    *str;
    {
    int     i;
    char    c;

    refresh();
    for (i = 0; (c=ttgetc()&0x7F) != '\r' && c != '\n' && c != '\000'; i++) {
	if (c == 0x7F || c == 0x8) {
	    if (col > 0) {col--; i--; }
	    addchr(' ');
	    col--;
	    i--;
	    }
	else {
	    addchr(c);
	    str[i] = c;
	    }
	refresh();
	if (i > 79) break;
	}
    str[i] = '\000';
    if (i == 0) return((char *)0L);
    else	return(str);
    }

ttgetc() {
#ifdef MCH_AMIGA
    unsigned char ch[2];

    if (!inited) return(0);
    Read(terminal, ch, 1L);
    return (ch[0] & 0xFF);
#endif

#ifdef VMS
    int	ch;

    if (!inited) return (0);
    term[0] = 0;
    term[1] = 0;
    status = SYS$QIOW(EFN, iochan,IO$_READLBLK|IO$M_NOECHO|IO$M_NOFILTR,
	iosb, 0, 0, &ch, 1, 0, term, 0, 0);
    if (status != SS$_NORMAL) exit(status);
    if (status != SS$_NORMAL || (status = (iosb[0]&0xFFFF)) != SS$_NORMAL)
	    exit(status);
    return ch & 0x7F;
#endif

#ifdef U__X
    if (!inited) return(0);
    return(getchar() & 0x7F);
#endif
    }

ttputc(c)
#ifdef MCH_AMIGA
char c;
#endif
    {
    if (!inited) return;
#ifdef MCH_AMIGA
    scrn_tmp[scrn_tmp_p++] = c;
    if(scrn_tmp_p>=AMG_MAXBUF) amg_flush();
#endif

#ifdef VMS
    if (nobuf >= NOBUF) ttflush();
    obuf[nobuf++] = c;
#endif

#ifdef U__X
    fputc(c, stdout);
#endif
    }

#ifdef MCH_AMIGA
amg_flush()
    {
    if(scrn_tmp_p) Write(terminal,scrn_tmp,(long)scrn_tmp_p);
    scrn_tmp_p = 0;
    }
#endif

ttputs(s)
char	*s;
    {
    while (*s) ttputc(*s++);
    }

ttflush()
    {
    if (!inited) return;
#ifdef MCH_AMIGA
    amg_flush();
#endif

#ifdef VMS
    status = SS$_NORMAL;
    if (nobuf != 0) {
	status = SYS$QIOW(EFN, iochan, IO$_WRITELBLK|IO$M_NOFORMAT,
	iosb, 0, 0, obuf, nobuf, 0, 0, 0, 0);
	if (status == SS$_NORMAL) status = iosb[0] & 0xFFFF;
	nobuf = 0;
	}
    return (status);
#endif

#ifdef U__X
    fflush(stdout);
#endif
    }


tabify(cp,dp)
char	*cp,*dp;
    {
    int	dcol,ocol;

    dcol = ocol = 0;
    while (1) {
	if (*cp == ' ') dcol++;
    	else {
	    while (((ocol + 8) &~ 07) <= dcol) {
		if (ocol + 1 == dcol) break;
		*dp++   = '\t';
		ocol   += 8;
		ocol   &= ~07;
		}
	    while (ocol < dcol) {
		*dp++	= ' ';
		ocol++;
		}
	    if (!*cp) {
		while (*--dp == ' ' || *dp == '\t') ;
		*++dp = '\000';
		return;
		}
	    if (*cp == '\033')
		for (*dp++ = *cp++; *cp == '[' || *cp < 'A'; *dp++ = *cp++) ;
	    else {
		ocol++;
		dcol++;
		}
	    *dp++ = *cp;
	    }
	cp++;
	}
    }


