/************************************************************************
 *									*
 *			Copyright (c) 1990 by     			*
 *		Digital Equipment Corporation, Maynard, MA		*
 *			All rights reserved.				*
 *									*
 *   This software is furnished under a license and may be used and	*
 *   copied  only  in accordance with the terms of such license and	*
 *   with the  inclusion  of  the  above  copyright  notice.   This	*
 *   software  or  any  other copies thereof may not be provided or	*
 *   otherwise made available to any other person.  No title to and	*
 *   ownership of the software is hereby transferred.			*
 *									*
 *   The information in this software is subject to change  without	*
 *   notice  and should not be construed as a commitment by Digital	*
 *   Equipment Corporation.						*
 *									*
 *   Digital assumes no responsibility for the use  or  reliability	*
 *   of its software on equipment which is not supplied by Digital.	*
 *									*
 *									*
 ************************************************************************
 *									*
 *			Modification History				*
 *									*
 ************************************************************************
 *									*
 * Authors:								*
 *	DBW - David B. Wecker, Database Systems Research Group	      	*
 *		DTN: 	523-2794					*
 *		PHONE:	719/260-2794					*
 *		ENET:	CHILD::WECKER					*
 *		USENET:	decwrl!child.dec.com!wecker			*
 *		ARPA:	wecker%child.dec.com@decwrl.dec.com		*
 *		MS: 	CXN/2 Upper - 564				*
 *		MAIL:	1175 Chapel Hills Drive, Col Spgs, CO 80920-3952*
 *									*
 * Versions:								*
 *  v1.0 DBW 871105 Ultrix/X10 based version				*
 *  v2.1 DBW 871117 Major re-write (non X based)			*
 *  v2.2 DBW 880114 Update of internal functions			*
 *  v3.0 DBW 881213 Major new release (replaced NeurDS compiler)	*
 *  v3.1 DBW 890501 Minor release for external use (portability fixes)	*
 *									*
 ************************************************************************/


/***************************************************************************
 *
 *  v1.0 DBW 881128 First version D. Wecker
 *    .1 DBW 881202 Added postscript and multi plot support
 *    .2 DBW 881208 Added Error information
 *    .3 DBW 881214 Added my own getopt() for the benefit of VMS 
 *
 ***************************************************************************/

#include "stdio.h"
#include "math.h"
#include "ctype.h"

char *usage = "Usage: plt [options] [ifile]\n\n\
options:\n\
    -a		average records in a ground\n\
    -c col	column to start scan at\n\
    -d idx	drop lowest bit from INENN field (for bias bit)\n\
    -f fmt	scanf format string (use %%f or %%s (INENN) for each entry)\n\
    -g cnt	group every n records (default == 1)\n\
    -l		take log of values\n\
    -n num	number of samples/ps page (default == all)\n\
    -p pfile	output postscript file\n\
    -o ofile	output file (default == stdout)\n\
    -r		show row number instead of plotted value as label\n\
    -s idx	sort by idx field\n\
    -t str	use str as title for output plot\n\
    -1 idx	idx of scanned number to use (starting from 0 (default last))\n\
    -2 idx	plot 2 fields (idx is index of 2nd item to plot)\n\
    ifile	input  file (default == stdin)\n\n";

char *pshead = "%!\n\
/inch { 72 mul } def /pagelen 11 inch def /pagewidth 8.5 inch def\n\
/xbase .5 inch def /ybase 10.5 inch def /ylen 10 inch def /xlen 7.5 inch def\n\
/ycur ybase def /xcur xbase def /numpts 100 def\n\
/ydecr { /ycur ycur yincr sub def } def\n\
/title { /titstr exch def gsave /Times-Bold findfont 14 scalefont setfont\n\
xbase 2 div pagelen 2 div titstr stringwidth pop 2 div add translate -90\n\
rotate 0 0 moveto titstr show grestore } def\n\
/init1 { /xcur exch def /numpts exch def /ycur ybase def /yincr ylen numpts\n\
div def /gsave 1 setlinewidth 1 setlinecap 1 setlinejoin [] 0 setdash\n\
newpath xbase xcur xlen mul add ycur moveto ydecr } def\n\
/init2 { init1 [2 3] 0 setdash } def\n\
/pt { xlen mul xbase add ycur lineto ydecr } def\n\
/fin { stroke grestore } def\n";

extern	char *optarg;
extern	int  optind;

#define	TOTMAX	10
#define tab(c)	((c+8) & ~7)

FILE    *ifp = NULL;
FILE	*ofp = NULL;
FILE	*pfp = NULL;
char	ifname[128] = "\0";
char	ofname[128] = "\0";
char	pfname[128] = "\0";
char	title[128]  = "\0";
char	fmt[128]    = "%lf";
int	idx = -1;
int	idx2= -1;
int	drop= -1;
int	col = 0;
int	tot = 0;
int	avg = 0;
int	grp = 1;
int	srt = -1;
int	spp = 0;
int	dorow = 0;
int	dolog = 0;
char	str[256];
char	scn[TOTMAX][128];
int	INENN[TOTMAX];
double	avg1 = 0.0,
	avg2 = 0.0,
	var1 = 0.0,
	var2 = 0.0,
	rms  = 0.0,
	err  = -999.9;

struct val {
    double  srt,val1,val2;
    } vals[10000];
double	maxval = -9e+30,minval = 9e+30;
int	vidx;

compar(v1,v2)
struct val *v1,*v2;
    {
    if	    (v1->srt <  v2->srt)    return -1;
    else if (v1->srt == v2->srt)    return  0;
    else			    return  1;
    }

leave(msg,a1,a2,a3,a4)
char	*msg,*a1,*a2,*a3,*a4;
    {
    if (pfp)				fclose(pfp);
    if (ifp && ifp != stdin)		fclose(ifp);
    if (ofp && ofp != stdout)		fclose(ofp);
    if (msg) {
	printf(msg,a1,a2,a3,a4);
	putchar('\n');
	exit(1);
	}
    exit(0);
    }

cvt(idx) {
    int	    i,j,pos,bit,cnt,val;
    int	    nibble;
    int	    final = 0,scr,shft = 0;
    double  *d;

    for (pos=strlen(scn[idx])-1; pos > -1 && !isxdigit(scn[idx][pos]); pos--) ;
    if (idx == drop) bit = 5;
    else	     bit = 4;
    pos++;
    while (pos > -1) {
	for (i = cnt = val = 0; i < 10; i++) {
	    if (bit >= 4) {
		pos--;
		bit -= 4;
		if (pos < 0 || !isxdigit(scn[idx][pos])) break;
		sscanf(&scn[idx][pos],"%01x",&nibble);
		}
	    j = (nibble >> bit++) & 1;
	    if (j) { cnt++; val += i-1; }
	    }
	if (cnt)    scr = (int)(((double)val / (double)cnt) + 0.5);
	else	    scr = 0;
	if (scr > 7)   scr = 7;
	if (scr < 0)   scr = 0;
	final += scr << shft;
	shft  += 3;
	}
    d  = (double *)scn[idx];
    *d = (double)final;
    }

main(argc,argv)
char **argv;
    {
    int	    i,j,k,l,cur,col;
    char    chr;
    double  v,v2,srtv;

    ifp = stdin;
    ofp = stdout;

    while ((chr=getopt_local(argc,argv,"ad:g:c:1:2:lrs:f:o:p:t:n:")) != EOF)
	switch (chr) {
	    case 'a':	avg++;					break;
	    case 'g':	grp  = atoi(optarg);			break;
	    case 'c':	col  = atoi(optarg);			break;
	    case 'd':	drop = atoi(optarg);			break;
	    case 'n':	spp  = atoi(optarg);			break;
	    case '1':	idx  = atoi(optarg);			break;
	    case '2':	idx2 = atoi(optarg);			break;
	    case 'l':	dolog++;				break;
	    case 'r':	dorow++;				break;
	    case 's':	srt  = atoi(optarg);			break;
	    case 't':	strcpy(title,optarg);			break;
	    case 'f':	strcpy(fmt,optarg);			break;
	    case 'o':	strcpy(ofname,optarg);
			ofp = fopen(ofname,"w");
			if (!ofp) leave("Can't open %s",ofname);
								break;
	    case 'p':	strcpy(pfname,optarg);
			pfp = fopen(pfname,"w");
			if (!pfp) leave("Can't open %s",pfname);
			fputs(pshead,pfp);
								break;
	    case '?':	leave(usage,NULL);			break;
	    }
    
    for (i=1; fmt[i]; i++)
	if (fmt[i-1] == '%' && fmt[i] >= 'a' && fmt[i] <= 'z') {
	    if (fmt[i] == 's')	INENN[tot] = 1;
	    else		INENN[tot] = 0;
	    tot++;
	    }
    if (idx < 0) idx = tot-1;
    if (idx >= tot) leave("Index greater than total %%s's or %%f's in format");
    if (srt >= tot) leave("Sort  greater than total %%s's or %%f's in format");
    if (idx2 >= tot) 
	leave("Index 2 greater than total %%s's or %%f's in format");
    if (avg && srt >=0) leave("Can't use sort and average together");
    if (grp < 1)	leave("Group must be at least 1");
    if (tot > TOTMAX)	leave("Total fields > %d",TOTMAX);
    if (optind < argc) {
	strcpy(ifname,argv[optind]);
	ifp = fopen(ifname,"r");
	if (!ifp) leave("Can't open %s",ifname);
	}
    if (++optind < argc) leave(usage);

    v = 0.0; v2 = 0.0; srtv = 0.0; cur = 0;
    while (fgets(str,256,ifp))
	if (sscanf(&str[col],fmt,scn[0],scn[1],scn[2],scn[3],scn[4],
		scn[5],scn[6],scn[7],scn[8],scn[9]) == tot) {
	    if (INENN[idx]) cvt(idx);
	    if (!avg) v  = 0.0;
	    v += *((double *)scn[idx]);
	    if (idx2 >= 0) {
		if (INENN[idx2] && idx2 != idx) cvt(idx2);
		if (!avg) v2  = 0.0;
		v2 += *((double *)scn[idx2]);
		}
	    if (srt >= 0) {
		if (INENN[srt] && srt != idx && srt != idx2) cvt(srt);
		if (!avg) srtv  = 0.0;
		srtv += *((double *)scn[srt]);
		}
	    if (++cur < grp) continue;
	    vals[vidx].val1 = v;
	    v		    = 0.0;
	    cur		    = 0;
	    if (avg)	vals[vidx].val1 /= (double)grp;
	    if (vals[vidx].val1 > maxval) maxval = vals[vidx].val1;
	    if (vals[vidx].val1 < minval) minval = vals[vidx].val1;
	    avg1 += vals[vidx].val1;
	    var1 += vals[vidx].val1 * vals[vidx].val1;
	    if (idx2 >= 0) {
		vals[vidx].val2	= v2;
		if (avg)	vals[vidx].val2 /= (double)grp;
		if (vals[vidx].val2 > maxval) maxval = vals[vidx].val2;
		if (vals[vidx].val2 < minval) minval = vals[vidx].val2;
		avg2 += vals[vidx].val2;
		var2 += vals[vidx].val2 * vals[vidx].val2;
		v2    = vals[vidx].val1 - vals[vidx].val2;
		rms  += v2 * v2;
		v2    = 0.0;
		}
	    if (srt >= 0) {
		vals[vidx].srt	= srtv;
		srtv		= 0.0;
		if (avg)	vals[vidx].srt /= (double)grp;
		}
	    vidx++;
	    }
    if (!vidx) leave("No points found");
    avg1 /= vidx;
    var1 /= vidx;
    var1 -= avg1 * avg1;
    if (idx2 >= 0) {
	avg2 /= vidx;
	var2 /= vidx;
	var2 -= avg2 * avg2;
	rms  /= vidx;
	rms   = sqrt(rms);
	}
    fprintf(ofp,"=========================================================\n");
    if (title[0]) fprintf(ofp,"%s\n",title);
    fprintf(ofp,"Samples: %4d, Minimum: %10.4e, Maximum: %10.4e\n",
	vidx,minval,maxval);
    fprintf(ofp,"Curve 1:       Mean:    %10.4e, SD:      %10.4e\n",
	avg1,sqrt(var1));
    if (idx2 >= 0) {
	fprintf(ofp,"Curve 2:       Mean:    %10.4e, SD:      %10.4e\n",
	    avg2,sqrt(var2));
	if (var2 == 0.0) var2 = 1.0;
	fprintf(ofp,"               RMS:     %10.4e, Err Idx: %10.4e\n",
	    rms,err=rms/sqrt(var1));
	}
    fprintf(ofp,"=========================================================\n");
    if (srt >= 0) qsort(vals,vidx,sizeof(struct val),compar);
    if (dolog) {
	maxval = log(maxval);
	minval = log(minval);
	}
    maxval -= minval;
    if (maxval == 0.0) maxval = 1.0;
    for (i=0; i<vidx; i++) {
	if (dolog) vals[i].val1 = log(vals[i].val1);
	vals[i].val1 -= minval;
	vals[i].val1 /= maxval;
	if (idx2 >= 0) {
	    if (dolog) vals[i].val2 = log(vals[i].val2);
	    vals[i].val2 -= minval;
	    vals[i].val2 /= maxval;
	    }
	}

    if (pfp) {
	if (!spp) spp = vidx;
	for (j=0,k=1; j<vidx; j += spp,k++) {
	    fprintf(pfp,"%d %.3lf init1\n",spp,vals[j].val1);
	    for (i=j+1; i<j+spp; i++) fprintf(pfp,"%.3lf pt\n",vals[i].val1);
	    fprintf(pfp,"fin\n");
	    if (idx2 >= 0) {
		fprintf(pfp,"%d %.3lf init2\n",spp,vals[j].val2);
		for (i=j+1; i<j+spp; i++) fprintf(pfp,"%.3lf pt\n",vals[i].val2);
		fprintf(pfp,"fin\n");
		}

	    if (err >= 0.0) sprintf(str,"%s [ERROR: rms %5.3lf, idx %5.3lf]",
		title,rms,err);
	    else	    strcpy(str,title);

	    fprintf(pfp,"(Page %d/%d:    %s) title\n",k,((vidx-1)/spp)+1,str);
	    fprintf(pfp,"showpage\n");
	    }
	fclose(pfp);
	pfp = NULL;
	}

    for (i=0; i<vidx; i++) {
	j = (int)(69.0 * vals[i].val1) + 1;
	if (idx2 >= 0)	k = (int)(69.0 * vals[i].val2) + 1;
	else		k = -1;

	if (dorow)	    fprintf(ofp,"%7d ",i);
	else if (srt >= 0)  fprintf(ofp,"%8.2e",vals[i].srt);
	else {
	    v = vals[i].val1 * maxval + minval;
	    if (dolog) v = exp(v);
	    fprintf(ofp,"%8.2e",v);
	    }
	col = 0;
	while (j > col || k > col) {
	    if (k > col && j > k) {
		while (k >  tab(col))	{ fputc('\t',ofp); col = tab(col); }
		while (k >  col)	{ fputc(' ',ofp);  col++; }
		fputc('2',ofp); 
		col++;
		if (j <= col) { fputc('1',ofp); col++; }
		}
	    else if (j > col && k >= j) {
		while (j >  tab(col))	{ fputc('\t',ofp); col = tab(col); }
		while (j >  col)	{ fputc(' ',ofp);  col++; }
		if (k == j) fputc('*',ofp);
		else	    fputc('1',ofp);
		col++;
		if (k != j && k <= col) { fputc('2',ofp); col++; }
		}
	    else {
		if (j > k) l = j;
		else	   l = k;
		while (l >  tab(col))	{ fputc('\t',ofp); col = tab(col); }
		while (l >  col)	{ fputc(' ',ofp);  col++; }
		if (j > k)  fputc('1',ofp);
		else	    fputc('2',ofp);
		col = l;
		}
	    }
	fputc('\n',ofp);
	}
    leave(NULL);
    }

/* got this off net.sources */
 
int	opterr = 1,		/* useless, never set or used */
	optind = 1,		/* index into parent argv vector */
	optopt;			/* character checked for validity */
char	*optarg;		/* argument associated with option */
 
#define BADCH	(int)'?'
#define EMSG	""
#define tell(s)	fputs(*nargv,stderr);fputs(s,stderr); \
		fputc(optopt,stderr);fputc('\n',stderr);return(BADCH);
 
getopt_local(nargc,nargv,ostr)
int	nargc;
char	**nargv,
	*ostr;
    {
    static char		*place = EMSG;	/* option letter processing */
    register char	*oli;		/* option letter list index */
    char		*strchr();
 
    if(!*place) {			/* update scanning pointer */
	if(optind >= nargc || *(place = nargv[optind]) != '-' || !*++place) return(EOF);
	if (*place == '-') {	/* found "--" */
	    ++optind;
	    return(EOF);
	    }
	}				/* option letter okay? */
    if ((optopt = (int)*place++) == (int)':' || !(oli = strchr(ostr,optopt))) {
	if(!*place) ++optind;
	tell(": illegal option -- ");
	}
    if (*++oli != ':') {		/* don't need argument */
	optarg = NULL;
	if (!*place) ++optind;
	}
    else {				/* need an argument */
	if (*place) optarg = place;	/* no white space */
	else if (nargc <= ++optind) {	/* no arg */
	    place = EMSG;
	    tell(": option requires an argument -- ");
	    }
	 else optarg = nargv[optind];	/* white space */
	place = EMSG;
	++optind;
	}
    return(optopt);			/* dump back option letter */
    }
