/************************************************************************
 *									*
 *			Copyright (c) 1990 by     			*
 *		Digital Equipment Corporation, Maynard, MA		*
 *			All rights reserved.				*
 *									*
 *   This software is furnished under a license and may be used and	*
 *   copied  only  in accordance with the terms of such license and	*
 *   with the  inclusion  of  the  above  copyright  notice.   This	*
 *   software  or  any  other copies thereof may not be provided or	*
 *   otherwise made available to any other person.  No title to and	*
 *   ownership of the software is hereby transferred.			*
 *									*
 *   The information in this software is subject to change  without	*
 *   notice  and should not be construed as a commitment by Digital	*
 *   Equipment Corporation.						*
 *									*
 *   Digital assumes no responsibility for the use  or  reliability	*
 *   of its software on equipment which is not supplied by Digital.	*
 *									*
 *									*
 ************************************************************************
 *									*
 *			Modification History				*
 *									*
 ************************************************************************
 *									*
 * Authors:								*
 *	DBW - David B. Wecker, Database Systems Research Group	      	*
 *		DTN: 	523-2794					*
 *		PHONE:	719/260-2794					*
 *		ENET:	CHILD::WECKER					*
 *		USENET:	decwrl!child.dec.com!wecker			*
 *		ARPA:	wecker%child.dec.com@decwrl.dec.com		*
 *		MS: 	CXN/2 Upper - 564				*
 *		MAIL:	1175 Chapel Hills Drive, Col Spgs, CO 80920-3952*
 *									*
 * Versions:								*
 *  v1.0 DBW 871105 Ultrix/X10 based version				*
 *  v2.1 DBW 871117 Major re-write (non X based)			*
 *  v2.2 DBW 880114 Update of internal functions			*
 *  v3.0 DBW 881213 Major new release (replaced NeurDS compiler)	*
 *  v3.1 DBW 890501 Minor release for external use (portability fixes)	*
 *									*
 ************************************************************************/


/*
 * srandom(x)	- sets a random seed starting point (long int)
 * random()	- returns next random long int (between 0 and 0x7FFFFFFF)
 * rnd()	- return next random double (between 0 and 1)
 * rndNorm()	- return a random normal with rndNormMean and rndNormVar
 */

double	rndNormMean 	= 0.5;
double  rndNormVar  	= 0.2;

#define rndDoubleScale	4.656612873077392578125e-10

static  long		randtbl[ 63 ]	= {
			0x00000001, 0x41c67ea6, 0x967eb0e7,
			0x2781e494, 0xc46b9b3d, 0xf94bdf32, 0x95fb7483,
			0xd9e2b600, 0x9cfbae39, 0xbf54bc7e, 0x0ff6d5df,
			0x0abd322c, 0x31dff4f5, 0x237c228a, 0xaf1cf0fb,
			0x7de14518, 0xc487eb71, 0xe201dd56, 0xd2bfa1d7,
			0xe2319ac4, 0xe3decdad, 0xe95678e2, 0x93728473,
			0x500f9f30, 0x284797a9, 0x2c67412e, 0x7566f4cf,
			0x42877e5c, 0xb3590565, 0x4daa423a, 0xe4880eeb,
			0xf73c2448, 0xef5992e1, 0xeeea4806, 0x5656aec7,
			0xe1133cf4, 0xca7b7c1d, 0x31d2de92, 0xad857063,
			0x99a13460, 0x69d8bd19, 0xb53c51de, 0x3b54afbf,
			0x7d55368c, 0x102f11d5, 0x1b37adea, 0x7aae88db,
			0xde652f78, 0xb45bf651, 0xe3dabeb6, 0x6102d7b7,
			0x7a79cb24, 0x8398a68d, 0x28cc1042, 0xe0e33853,
			0x9f1a7590, 0x4b761e89, 0x858eee8e, 0xfa5f06af,
			0x28d95abc, 0x87991a45, 0x388f659a, 0x751f5ecb };

static  long		*fptr			= &randtbl[ 2 ];
static  long		*rptr			= &randtbl[ 0 ];
static  long		*end_ptr		= &randtbl[ 63 ];

srandom( x )
unsigned long x;
    {
    register  int i;

    randtbl[ 0 ] = x;
    for( i = 1; i < 63; i++ )
	randtbl[i] = 1103515245*randtbl[i - 1] + 12345;

    fptr = &randtbl[ 1 ];
    rptr = &randtbl[ 0 ];
    }

long random()
    {
    long		i;
	
    *fptr += *rptr;
    i = (*fptr >> 1)&0x7fffffffL;	/* chucking least random bit */
    if(  ++fptr  >=  end_ptr  )  {
	fptr = randtbl;
	++rptr;
	}
    else  if(  ++rptr  >=  end_ptr  )  rptr = randtbl;
    return( i );
    }

double rnd() {
    return ((double)random()) * rndDoubleScale;
    }

double rndNorm()
{
    static int 	    cached = 0;
    static double   x2;
    double  	    u1,u2,v1,v2,w,y,x1;
    double  	    sqrt(),log();

    if (cached) {
	cached = 0;
	return (x2 * rndNormVar + rndNormMean);
	}
    else {
	while (1) {
	    u1	= rnd();
	    u2  = rnd();
	    v1	= 2. * u1 - 1.;
	    v2  = 2. * u2 - 1.;
	    w   = (v1 * v1) + (v2 * v2);
	    if (w <= 1. && w > 0.0) {
		y  = sqrt((-2. * log(w)) / w);
		x1 = v1 * y;
		x2 = v2 * y;
		cached = 1;
		return (x1 * rndNormVar + rndNormMean);
		}
	    }
	}
    }
